"use strict";

function _typeof(obj) { if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; var ownKeys = Object.keys(source); if (typeof Object.getOwnPropertySymbols === 'function') { ownKeys = ownKeys.concat(Object.getOwnPropertySymbols(source).filter(function (sym) { return Object.getOwnPropertyDescriptor(source, sym).enumerable; })); } ownKeys.forEach(function (key) { _defineProperty(target, key, source[key]); }); } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } }

function _createClass(Constructor, protoProps, staticProps) { if (protoProps) _defineProperties(Constructor.prototype, protoProps); if (staticProps) _defineProperties(Constructor, staticProps); return Constructor; }

function _possibleConstructorReturn(self, call) { if (call && (_typeof(call) === "object" || typeof call === "function")) { return call; } return _assertThisInitialized(self); }

function _getPrototypeOf(o) { _getPrototypeOf = Object.setPrototypeOf ? Object.getPrototypeOf : function _getPrototypeOf(o) { return o.__proto__ || Object.getPrototypeOf(o); }; return _getPrototypeOf(o); }

function _assertThisInitialized(self) { if (self === void 0) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function"); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, writable: true, configurable: true } }); if (superClass) _setPrototypeOf(subClass, superClass); }

function _setPrototypeOf(o, p) { _setPrototypeOf = Object.setPrototypeOf || function _setPrototypeOf(o, p) { o.__proto__ = p; return o; }; return _setPrototypeOf(o, p); }

var EventEmitter = require("events");

var _ = require("lodash");

var _require = require("../constants"),
    ERR_CODES = _require.ERR_CODES,
    ERR_MSGS = _require.ERR_MSGS;
/**
 * @class Client
 * @extends require('events').EventEmitter
 * @param {Object} [server] object with host and port of server
 * @param {Object} [options]
 * @param {Number} [options.version=2] JSON-RPC version to use (1|2)
 * @param {String} [options.delimiter="\n"] delimiter to use for requests
 * @param {Number} [options.timeout=30] timeout for request
 * @return {Client}
 */


var Client =
/*#__PURE__*/
function (_EventEmitter) {
  _inherits(Client, _EventEmitter);

  function Client(options) {
    var _this;

    _classCallCheck(this, Client);

    _this = _possibleConstructorReturn(this, _getPrototypeOf(Client).call(this));

    if (!(_assertThisInitialized(_this) instanceof Client)) {
      return _possibleConstructorReturn(_this, new Client(options));
    }

    _this.client = undefined;
    _this.message_id = 1;
    _this.serving_message_id = 1;
    _this.pendingCalls = {};
    _this.pendingBatches = {};
    _this.attached = false;
    /**
     * we can receive whole messages, or parital so we need to buffer
     *
     * whole message: {"jsonrpc": 2.0, "params": ["hello"], id: 1}
     *
     * partial message: {"jsonrpc": 2.0, "params"
     */

    _this.messageBuffer = "";
    _this.responseQueue = {};
    _this.options = options;
    _this.options.timeout = _this.options.timeout * 1000;
    var _this$options = _this.options,
        host = _this$options.host,
        port = _this$options.port;
    _this.server = {
      host: host,
      port: port
    };
    return _this;
  }

  _createClass(Client, [{
    key: "end",
    value: function end() {
      var _this2 = this;

      return new Promise(function (resolve, reject) {
        _this2.client.end(function (error) {
          if (error) {
            reject();
          }

          resolve();
        });
      });
    }
  }, {
    key: "initClient",
    value: function initClient() {
      this.client = new WebSocket(this.options); // need overrides for ws functions

      this.client.send = this.client.write;
    }
  }, {
    key: "connect",
    value: function connect() {
      var _this3 = this;

      return new Promise(function (resolve, reject) {
        _this3.client.onopen = function (event) {
          _this3.listen();

          _this3.handleResponse();

          _this3.handleError();

          resolve(event);
        };
      });
    }
  }, {
    key: "listen",
    value: function listen() {
      var _this4 = this;

      this.writer.onmessage = function (_ref) {
        var data = _ref.data;
        _this4.messageBuffer += data;

        var messages = _this4.messageBuffer.split(_this4.options.delimiter);

        if (messages.length > 1) {
          // otherwise messages are still coming in and request
          // possibly hasnt finished
          _this4.messageBuffer = "";

          _this4.verifyData(messages);
        }
      };

      this.client.onclose = function () {
        _this4.attached = false;

        _this4.client.removeAllListeners();

        _this4.emit("serverDisconnected");
      };
    }
  }, {
    key: "request",
    value: function request() {
      var _this5 = this;

      return {
        message: function message(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            id: _this5.message_id,
            options: _this5.options
          });
          _this5.message_id += 1;
          return request;
        },
        send: function send(method, params) {
          return new Promise(function (resolve, reject) {
            var requestId = _this5.message_id;
            _this5.pendingCalls[requestId] = {
              resolve: resolve,
              reject: reject
            };

            var request = _this5.request().message(method, params);

            _this5.options.headers["Content-Length"] = Buffer.byteLength(request, "utf-8");

            _this5.client.write(request);

            setTimeout(function () {
              if (_this5.pendingCalls[requestId]) {
                var error = _this5.sendError({
                  id: requestId,
                  code: ERR_CODES.timeout,
                  message: ERR_MSGS.timeout
                });

                delete _this5.pendingCalls[requestId];

                _this5.client.end();

                reject(error);
              }
            }, _this5.options.timeout);
          });
        },

        /**
         * The spec for HTTP notifications states a 204 error response with an empty body
         * Want to provide a promise response for users to know notification was received
         * or rejected by server
         */
        notify: function notify(method, params) {
          var request = formatRequest({
            method: method,
            params: params,
            options: _this5.options
          });
          return new Promise(function (resolve, reject) {
            _this5.options.headers["Content-Length"] = Buffer.byteLength(request, "utf-8");
            var notification = http.request(_this5.options, function (response) {
              if (response.statusCode === 204) {
                resolve(response);
              } else {
                reject(new Error("no response receieved for notification"));
              }
            });
            notification.write(request);
            notification.end();
          });
        }
      };
    }
  }, {
    key: "batch",
    value: function batch(requests) {
      var _this6 = this;

      /**
       * should receive a list of request objects
       * [client.request.message(), client.request.message()]
       * send a single request with that, server should handle it
       *
       * We want to store the IDs for the requests in the batch in an array
       * Use this to reference the batch response
       * The spec has no explaination on how to do that, so this is the solution
       */
      return new Promise(function (resolve, reject) {
        var batchIds = [];
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = requests[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var _request = _step.value;
            var json = JSON.parse(_request);

            if (json.id) {
              batchIds.push(json.id);
            }
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator["return"] != null) {
              _iterator["return"]();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }

        _this6.pendingBatches[String(batchIds)] = {
          resolve: resolve,
          reject: reject
        };
        var request = JSON.stringify(requests);
        _this6.options.headers["Content-Length"] = Buffer.byteLength(request, "utf-8");

        _this6.client.write(request);

        _this6.on("batchResponse", function (batch) {
          var batchResponseIds = [];
          batch.forEach(function (message) {
            if (message.error) {
              // reject the whole message if there are any errors
              reject(batch);
            }

            if (message.id) {
              batchResponseIds.push(message.id);
            }
          });

          if (_.isEmpty(batchResponseIds)) {
            resolve([]);
          }

          for (var _i = 0, _Object$keys = Object.keys(_this6.pendingBatches); _i < _Object$keys.length; _i++) {
            var ids = _Object$keys[_i];

            if (_.isEmpty(_.difference(JSON.parse("[".concat(ids, "]")), batchResponseIds))) {
              _this6.pendingBatches[ids].resolve(batch);
            }
          }
        });

        _this6.on("batchError", function (error) {
          reject(error);
        });
      });
    }
  }, {
    key: "handleResponse",
    value: function handleResponse() {
      var _this7 = this;

      this.on("response", function (id) {
        if (!(_this7.pendingCalls[id] === undefined)) {
          var response = _this7.responseQueue[id];

          if (_this7.writer instanceof http.IncomingMessage) {
            // want to allow users to access the headers, status code etc.
            response = _objectSpread({
              body: _this7.responseQueue[id]
            }, _this7.writer);
          }

          _this7.pendingCalls[id].resolve(response);

          delete _this7.responseQueue[id];
        }
      });
    }
  }, {
    key: "verifyData",
    value: function verifyData(messages) {
      var _iteratorNormalCompletion2 = true;
      var _didIteratorError2 = false;
      var _iteratorError2 = undefined;

      try {
        for (var _iterator2 = messages[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
          var chunk = _step2.value;

          try {
            // sometimes split messages have empty string at the end
            // just ignore it
            if (chunk !== "") {
              // will throw an error if not valid json
              var message = JSON.parse(chunk);

              if (_.isArray(message)) {
                // possible batch request
                try {
                  this.emit("batchResponse", message);
                } catch (e) {
                  var error = this.sendError({
                    id: this.serving_message_id,
                    code: ERR_CODES.parseError,
                    message: ERR_MSGS.parseError
                  });
                  this.emit("batchError", error);
                }
              }

              if (!_.isObject(message)) {
                // error out if it cant be parsed
                var _error = this.sendError({
                  id: null,
                  code: ERR_CODES.parseError,
                  message: ERR_MSGS.parseError
                });

                this.emit("messageError", _error);
              }

              if (!message.id) {
                // no id, so assume notification
                this.emit("notify", message);
              }

              if (message.error) {
                // got an error back so reject the message
                var _error2 = this.sendError({
                  jsonrpc: message.jsonrpc,
                  id: message.id,
                  code: message.error.code,
                  message: message.error.message
                });

                this.emit("messageError", _error2);
              } // no method, so assume response


              if (!message.method) {
                this.serving_message_id = message.id;
                this.responseQueue[this.serving_message_id] = message;
                this.emit("response", this.serving_message_id);
              }
            }
          } catch (e) {
            var _error3 = this.sendError({
              id: this.serving_message_id,
              code: ERR_CODES.parseError,
              message: ERR_MSGS.parseError
            });

            this.emit("messageError", _error3);
          }
        }
      } catch (err) {
        _didIteratorError2 = true;
        _iteratorError2 = err;
      } finally {
        try {
          if (!_iteratorNormalCompletion2 && _iterator2["return"] != null) {
            _iterator2["return"]();
          }
        } finally {
          if (_didIteratorError2) {
            throw _iteratorError2;
          }
        }
      }
    }
  }, {
    key: "listen",
    value: function listen() {
      var _this8 = this;

      this.writer.on("data", function (data) {
        _this8.messageBuffer += data;

        var messages = _this8.messageBuffer.split(_this8.options.delimiter);

        if (messages.length > 1) {
          // otherwise messages are still coming in and request
          // possibly hasnt finished
          _this8.messageBuffer = "";

          _this8.verifyData(messages);
        }
      });
      this.writer.on("end", function () {
        _this8.attached = false;

        _this8.client.removeAllListeners();

        _this8.emit("serverDisconnected");
      });
      this.client.on("close", function () {
        _this8.attached = false;

        _this8.client.removeAllListeners();
      });
    }
  }, {
    key: "serverDisconnected",
    value: function serverDisconnected(cb) {
      this.on("serverDisconnected", function () {
        return cb();
      });
    }
  }, {
    key: "handleError",
    value: function handleError() {
      var _this9 = this;

      this.on("messageError", function (error) {
        var response = error;

        if (_this9.writer instanceof http.IncomingMessage) {
          // want to allow users to access the headers, status code etc.
          response = _objectSpread({
            body: error
          }, _this9.writer);
        }

        _this9.pendingCalls[error.id].reject(response);
      });
    }
  }, {
    key: "sendError",
    value: function sendError(_ref2) {
      var jsonrpc = _ref2.jsonrpc,
          id = _ref2.id,
          code = _ref2.code,
          message = _ref2.message;
      var response;

      if (this.options.version === "2.0") {
        response = {
          jsonrpc: jsonrpc || this.options.version,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      } else {
        response = {
          result: null,
          error: {
            code: code,
            message: message || "Unknown Error"
          },
          id: id
        };
      }

      return response;
    }
  }]);

  return Client;
}(EventEmitter);

module.exports = Client;